/*
 * USB SQ905 driver - 0.2
 *
 * Copyright (C) 2005 Christophe Lucas <clucas@rotomalug.org>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/videodev.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>

#include "sqcam.h"

#define DEBUG

#define SEND	1
#define RECEIVE	0

static int forcev4l1; /* = 0 */
module_param(forcev4l1, int, 0644);
MODULE_PARM_DESC(forcev4l1, "force use of V4L1 instead of V4L2");

/* table of devices that work with this driver */
static struct usb_device_id sqcam_table [] = {
	{ USB_DEVICE(USB_SQCAM_VENDOR_ID, USB_SQCAM_SQ905_SENSOR_ID) },
	{ USB_DEVICE(USB_SQCAM_VENDOR_ID, USB_SQCAM_HV7131R_SENSOR_ID) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, sqcam_table);


/* Get a minor range for your devices from the usb maintainer */
#define USB_SQCAM_MINOR_BASE	-1
/* #define USB_SQCAM_MINOR_BASE	192 */


/*static struct sqcam sqcam_camera;*/

/* from bayer.c */
void sqcam_decode_bayer(const u8 *data, u8 *rgb, struct sqcam *cam);

static int sqcam_open(struct inode *inode, struct file *file);
static int sqcam_release(struct inode *inode, struct file *file);
static ssize_t sqcam_ioctl(struct inode *inode, struct file *file, unsigned int ioctlnr, unsigned long ul_arg);
static ssize_t sqcam_mmap(struct file *file, struct vm_area_struct *vma);
static ssize_t sqcam_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos);
static ssize_t sqcam_write(struct file *file, const char __user *user_buffer, size_t count, loff_t *ppos);
static void sqcam_disconnect(struct usb_interface *interface);
static int sqcam_probe(struct usb_interface *interface, const struct usb_device_id *id);
// static void sqcam_delete(struct kref *kref);


static struct file_operations sqcam_fops = {
	.owner =	THIS_MODULE,
	.read =		sqcam_read,
	.write =	sqcam_write,
	.open =		sqcam_open,
	.release =	sqcam_release,
	.ioctl = 	sqcam_ioctl,
	.mmap =		sqcam_mmap,
	.llseek	=	no_llseek,
};

static struct video_device sqcam_video_device = {
        .owner          = THIS_MODULE,
	.name           = "SQCam-based USB Camera",
	.type           = VID_TYPE_CAPTURE,
	.hardware       = VID_HARDWARE_SE401, /* need an own value */
	.release        = video_device_release,
	.fops           = &sqcam_fops,
	.minor          = -1,
};

static struct usb_driver sqcam_usb_driver = {
	.owner = THIS_MODULE,
	.name = "sqcam",
	.id_table = sqcam_table,
	.probe = sqcam_probe,
	.disconnect = sqcam_disconnect,
};

#define to_sqcam_dev(d) container_of(d, struct sqcam, kref)

/* 
 * useful functions
 * These are the same as stv680 driver and others.
 */
static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

unsigned long usbvideo_kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE-1); /* restore the offset */
	ret = __pa(kva);
	return ret;
}

static int sqcam_write_reg(struct sqcam* cam, u8 request, u16 value, u16 index, 
			unsigned char* tb, u16 size )
{
	int status;
	struct usb_device* udev = cam->udev ;
	
	unsigned char* t_buf = kmalloc(size, GFP_KERNEL);
	if (t_buf == NULL) {
		dev_err(&udev->dev, "memory allocation failed\n" );
		return -ENOMEM ;
	}
	memcpy(t_buf, tb, size);

	down(&cam->cam_lock);

	status = usb_control_msg(udev,usb_sndctrlpipe(udev, 0),request,
				USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE, 
				value, index,t_buf, size, HZ);
	if (status < 0) {
		dev_info(&udev->dev, "Failed to write a register (index 0x%02X, error %d)",
				index, status );
		return -1;
	}

	up(&cam->cam_lock);

	kfree(t_buf);
	cam->reg[index]=value;
	return 0;
}

static int sqcam_read_reg(struct sqcam* cam, u8 request, u16 value, u16 index, 
			unsigned char* tb, u16 size )
{
	int status;
	struct usb_device* udev = cam->udev ;
	
	unsigned char* t_buf = kmalloc(size, GFP_KERNEL );
	if (t_buf == NULL) {
		dev_err(&udev->dev, "memory allocation failed\n" );
		return -ENOMEM ;
	}
	memcpy(t_buf, tb, size);

	down(&cam->cam_lock);	

	status = usb_control_msg(udev,usb_rcvctrlpipe(udev, 0),request,
				USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE, 
				value, index,t_buf, size, HZ);
	if (status < 0) {
		dev_info(&udev->dev, "Failed to read a register (index 0x%02X, error %d)",
				index, status );
		return -1;
	}

	up(&cam->cam_lock);

	kfree(t_buf);
	return 0;

}


static int send_control_msg(int set, struct sqcam *cam, u8 request, u16 value,
		 u16 index, unsigned char *cp, u16 size)
{
	int status;

	// for reasons not yet known to me, you can't send USB control messages
	// with data in the module (if you are compiled as a module).  Whatever
	// the reason, copying it to memory allocated as kernel memory then
	// doing the usb control message fixes the problem.

	struct usb_device* udev = cam->udev;
	unsigned char* transfer_buffer = kmalloc(size, GFP_KERNEL);
	if( NULL == transfer_buffer ) {
		dev_dbg(udev->dev, "memory allocation faild :-(\n");
		return -ENOMEM ;
	}
	memcpy(transfer_buffer, cp, size);
	
	/*down(&cam->busy_lock);*/

	status = usb_control_msg(udev,
				 set ? usb_sndctrlpipe(udev, 0) : usb_rcvctrlpipe(udev, 0),
				 request,
				 (set ? USB_DIR_OUT : USB_DIR_IN) | USB_TYPE_VENDOR |
				 USB_RECIP_DEVICE, value, index,
				 transfer_buffer, size, HZ);

	kfree(transfer_buffer);

	dev_dbg(&udev->dev, "ctrl msg:: value: %0x, index: %0x, size: %i\n", value, index, size);
	status = min(status, 0);
	if (status < 0) {
		dev_err(&udev->dev, "send_msg: Failed sending control message, error %d.\n", status);
		dev_err(&udev->dev,  "ctrl msg:: dir: %i, value: %0x, index: %0x, size: %i\n", 
					set, value, index,  size
				);
	}
	/*up(&cam->busy_lock);*/

	return status;
}



static void sqcam_delete(struct kref *kref)
{	
	struct sqcam *dev = to_sqcam_dev(kref);

	if (dev) {
		if (dev->is_initialized)
			dev->sensor->reset_cam(dev);
		if (dev->vdev)
			video_unregister_device(dev->vdev);
		if (dev->udev)
			usb_put_dev(dev->udev);
		if (dev->bulk_in_buffer)
			kfree (dev->bulk_in_buffer);
		kfree (dev);
	}
}

static int sqcam_open(struct inode *inode, struct file *file)
{
	struct sqcam *dev;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;
	
	subminor = iminor(inode);

	interface = usb_find_interface(&sqcam_usb_driver, subminor);
	if (!interface) {
		err ("%s - error, can't find device for minor %d",
		     __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) {
		retval = -ENODEV;
		goto exit;
	}
	
	/* increment our usage count for the device */
	kref_get(&dev->kref);

	dev_dbg(&dev->udev->dev,"Avant de faire le init().\n");
	
	down(&dev->cam_lock);
	if (!dev->is_initialized) {
		dev->sensor->init_cam(dev);
		dev->is_initialized = 1 ;
	}
	up(&dev->cam_lock);

	/* save our object in the file's private structure */
	file->private_data = dev;

exit:
	return retval;
}

static int sqcam_release(struct inode *inode, struct file *file)
{
	struct sqcam *dev;

	dev = (struct sqcam *)file->private_data;
	if (dev == NULL)
		return -ENODEV;

	/* decrement the count on our device */
	kref_put(&dev->kref, sqcam_delete);
	return 0;
}

static ssize_t sqcam_ioctl(struct inode *inode, struct file *file, unsigned int ioctlnr, unsigned long ul_arg)
{
	return 0;
}

static ssize_t sqcam_mmap(struct file *file, struct vm_area_struct *vma)
{

	return 0;
}

static void read_frame(struct sqcam *cam, int framenum)
{
	unsigned char siz, msg;
	int n, read;
	int actual_length;

	if (!cam->udev) {
		return;
	}
	msg = 0x03; 
	read = 320 * 240 + 64;
	// read 'read' bytes in SQCAM_MAX_READ_SIZE chunks
		
	do {
		siz = 0x50;
		n = send_control_msg(SEND , cam, 0x0c , 0x03, 
				(read > SQCAM_MAX_READ_SIZE) ? SQCAM_MAX_READ_SIZE : read, 
				&siz, 1 );
		if (n < 0) {
			printk(KERN_ERR
		       		" sqcam_read_frame: Problem sending frame capture control message");
			goto done;
		}
		n = usb_bulk_msg( cam->udev,
			 usb_rcvbulkpipe(cam->udev, cam->bulk_in_endpointAddr),
			 cam->raw_image + (320*240+64 - read),
			 (read > SQCAM_MAX_READ_SIZE) ? SQCAM_MAX_READ_SIZE : read,
			 &actual_length, 500);
		read -= actual_length;
		dev_dbg(cam->udev->dev, "blk msg:: size: %d\n", actual_length);
		if (n < 0) {
			dev_err(&cam->udev->dev,"Problem during bulk read of frame data: %d\n", n);
			read = 0;
		}
	} while ( read );	
	
	if (forcev4l1 != 0) {
		sqcam_decode_bayer(cam->raw_image, 
			cam->framebuf + framenum * SQCAM_MAX_FRAME_SIZE,
			cam );
	}
	
	cam->framebuf_size =
	    320 * 240 * SQCAM_BYTES_PER_PIXEL;
	cam->framebuf_read_start = 0;
 done: 
	dev_dbg(&cam->udev->dev, "sqcam_read_frame: exiting \n");
}


static ssize_t sqcam_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
	struct sqcam *dev;
	int retval = 0;

	dev = (struct sqcam *)file->private_data;
	
	/* do a blocking bulk read to get data from the device */
	retval = usb_bulk_msg(dev->udev,
			      usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpointAddr),
			      dev->bulk_in_buffer,
			      min(dev->bulk_in_size, count),
			      &count, HZ*10);

	/* if the read was successful, copy the data to userspace */
	if (!retval) {
		if (copy_to_user(buffer, dev->bulk_in_buffer, count))
			retval = -EFAULT;
		else
			retval = count;
	}

	return retval;
}

static void sqcam_write_bulk_callback(struct urb *urb, struct pt_regs *regs)
{
	/* sync/async unlink faults aren't errors */
	if (urb->status && 
	    !(urb->status == -ENOENT || 
	      urb->status == -ECONNRESET ||
	      urb->status == -ESHUTDOWN)) {
		dbg("%s - nonzero write bulk status received: %d",
		    __FUNCTION__, urb->status);
	}

	/* free up our allocated buffer */
	usb_buffer_free(urb->dev, urb->transfer_buffer_length, 
			urb->transfer_buffer, urb->transfer_dma);
}

static ssize_t sqcam_write(struct file *file, const char __user *user_buffer, size_t count, loff_t *ppos)
{
	struct sqcam *dev;
	int retval = 0;
	struct urb *urb = NULL;
	char *buf = NULL;

	dev = (struct sqcam *)file->private_data;

	/* verify that we actually have some data to write */
	if (count == 0)
		goto exit;

	/* create a urb, and a buffer for it, and copy the data to the urb */
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!urb) {
		retval = -ENOMEM;
		goto error;
	}

	buf = usb_buffer_alloc(dev->udev, count, GFP_KERNEL, &urb->transfer_dma);
	if (!buf) {
		retval = -ENOMEM;
		goto error;
	}
	if (copy_from_user(buf, user_buffer, count)) {
		retval = -EFAULT;
		goto error;
	}

	/* initialize the urb properly */
	usb_fill_bulk_urb(urb, dev->udev,
			  usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr),
			  buf, count, sqcam_write_bulk_callback, dev);
	urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	/* send the data out the bulk port */
	retval = usb_submit_urb(urb, GFP_KERNEL);
	if (retval) {
		err("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
		goto error;
	}

	/* release our reference to this urb, the USB core will eventually free it entirely */
	usb_free_urb(urb);

exit:
	return count;

error:
	usb_buffer_free(dev->udev, count, buf, urb->transfer_dma);
	usb_free_urb(urb);
	kfree(buf);
	return retval;
}



/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with devfs and the driver core
 */
static struct usb_class_driver sqcam_class = {
	.name = "usb/sqcam%d",
	.fops = &sqcam_fops,
	.mode = S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH,
	.minor_base = USB_SQCAM_MINOR_BASE,
};

static int sqcam_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct sqcam *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
	int retval = -ENOMEM;

	/* allocate memory for our device state and initialize it */
	dev = kmalloc(sizeof(struct sqcam), GFP_KERNEL);
	if (dev == NULL) {
		dev_err(&dev->udev->dev,"Out of memory");
		goto error;
	}
	memset(dev, 0x00, sizeof (*dev));

	init_MUTEX(&dev->cam_lock);

	kref_init(&dev->kref);
	dev->is_initialized = 0;

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;

	if (dev->udev->descriptor.idVendor == USB_SQCAM_VENDOR_ID) {
		int i, ret ;
		for (i=0 ; sensor_table[i] ; i++ ) {
			ret = sensor_table[i](dev);
			if (!ret)
				break;
			
		}
	} else {
		dev_info(&dev->udev->dev,"No SQcam device found.\n");
		return -ENODEV;
	}
	dev_info(&dev->udev->dev, "SQcam based webcam connected.\n");
	

	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!dev->bulk_in_endpointAddr &&
		    (endpoint->bEndpointAddress & USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk in endpoint */
			buffer_size = endpoint->wMaxPacketSize;
			dev->bulk_in_size = buffer_size;
			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
			dev->bulk_in_buffer = kmalloc(buffer_size, GFP_KERNEL);
			if (!dev->bulk_in_buffer) {
				err("Could not allocate bulk_in_buffer");
				goto error;
			}
		}

		if (!dev->bulk_out_endpointAddr &&
		    !(endpoint->bEndpointAddress & USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
		}
	}
	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
		err("Could not find both bulk-in and bulk-out endpoints");
		goto error;
	}

	dev->vdev = video_device_alloc();
	if (!dev->vdev) {
		retval = -ENOMEM;
		goto error;
	}
	
	memcpy(dev->vdev, &sqcam_video_device, sizeof (sqcam_video_device));
        /* sort of a reverse mapping for those functions that get vdev only */
	video_set_drvdata(dev->vdev, dev);

	if (video_register_device(dev->vdev, VFL_TYPE_GRABBER, -1) == -1) {
		printk(KERN_WARNING "video_register_device failed\n");
		goto error_vdev;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* let the user know what node this device is now attached to */
	dev_info(&dev->udev->dev,"USB SQCam device now attached to USB-%d\n", interface->minor);
	return 0;

error_vdev:
	if (dev)
		video_device_release(dev->vdev);
error:
	if (dev) {
		kref_put(&dev->kref, sqcam_delete);
		kfree(dev);
	}
	return retval;
}

static void sqcam_disconnect(struct usb_interface *interface)
{
	struct sqcam *dev;
	int minor = interface->minor;

	dev = usb_get_intfdata(interface);

	if (dev==NULL) {
		printk( KERN_ERR "We have a little problem captain, device is NULL :-(\n");
	}

	/* prevent sqcam_open() from racing sqcam_disconnect() */
	dev_info(&dev->udev->dev, "USB SQCam #%d now disconnected.\n", minor);
	down(&dev->cam_lock);

	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);
	
	/* give back our minor */
//	usb_deregister_dev(interface, &sqcam_class); 
	dev->udev = NULL ;

	up(&dev->cam_lock);

	/* decrement our usage count */
	kref_put(&dev->kref, sqcam_delete);
}

static int __init usb_sqcam_init(void)
{
	int result;

	/* register this driver with the USB subsystem */
	result = usb_register(&sqcam_usb_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	return result;
}

static void __exit usb_sqcam_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&sqcam_usb_driver);
}

module_init (usb_sqcam_init);
module_exit (usb_sqcam_exit);

MODULE_AUTHOR("Christophe Lucas <clucas@rotomalug.org>");
MODULE_DESCRIPTION("v4l/v4l2 driver for the SQCam camera");
MODULE_VERSION(SQCAM_DRIVER_VERSION);
MODULE_LICENSE("GPL");
